// =*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*
//  
//   A Winbatch extender to return date and time information 
//   in any user-specified format.
//
// ---------------------------------------------------------------
//   includes the following routines
//
//       NowFormat(format)
//       TimeFormat(format,YmdHms)
//
// ----------------------------------------------------------------

#define EXTENDERVERSION 11106          // DLL version number
#define STRICT                         // only if you prototype carefully

/*-------------------------------------------------------------------------
!                                    Includes
+----------------------------------------------------------------------- */

                           // Standard Windows include
#include <windows.h>
                                  
                          // Special includes for WinBatch extender
#include "wwwpform.h"
#include "addons.h"

/*------------------------------------------------------------------------+
!                            Defines                                      !
+------------------------------------------------------------------------*/

     /*************************************************************
    !   LoadStr loads a string from the resource table.
    !   p1 = the Resource number
    **************************************************************/
#define LoadStr(p1)  LoadString(hinstDLL, (p1), szWork, sizeof(szWork))

    /*************************************************************
    !  RETLONG and RETSTRING pass their argument backs 
    !  to WinBatch, and exit the extender.
    **************************************************************/

#define RETLONG(p1)      return ((*lpPhoneHome)                            \
                                 ((int)   RESULT_SUCCESS,                  \
                                  (LONG)  lVersion,                        \
                                  (LPSTR) lpBlock,                         \
                                  (BOOL)  RESULT_LONG,                     \
                                  (LONG)  (p1),                            \
                                  (LPSTR) NULL)))

#define RETSTRING(p1)    return ((*lpPhoneHome)                            \
                                 ((int)   RESULT_SUCCESS,                  \
                                  (LONG)  lVersion,                        \
                                  (LPSTR) lpBlock,                         \
                                  (BOOL)  RESULT_STRING,                   \
                                  (LONG)  0,                               \
                                  (LPSTR) (p1)))

    /**************************************************************
    *  RETERROR(level,Err#) passes back an error massage 
    *  to WinBatch. Level MUST be -1, -2 or -3.
    *               -1 for most run of the mill errors.
    *               -2 for serious errors
    *               -3 for absolutely fatal "we better die" errors
    *  Err# is the code number of the error message on the RC file.
    **************************************************************/
#define RETERROR(p1, p2) return ((*lpPhoneHome)                            \
                                 ((int)   (p1),                            \
                                  (LONG)  lVersion,                        \
                                  (LPSTR) lpBlock,                         \
                                  (BOOL)  RESULT_STRING,                   \
                                  (LONG)  LoadStr(p2),                     \
                                  (LPSTR) (szWork)))
                                  
    /**************************************************************
    *   RETCANCEL returns a Cancel flag
    **************************************************************/
#define RETCANCEL       return -5

    /**************************************************************
    *  ANSWERSTRING passes its string argument back to Winbatch,
    *  but does not exit the extender.  This gives you a chance to
    *  do cleanup functions after returning the value.  The 
    *  return value from ANSWERSTRING must be saved, and passed
    *  as the return value when you do exit the function.
    **************************************************************/
#define ANSWERSTRING(p1)        ((*lpPhoneHome)                            \
                                 ((int)   RESULT_SUCCESS,                  \
                                  (LONG)  lVersion,                        \
                                  (LPSTR) lpBlock,                         \
                                  (BOOL)  RESULT_STRING,                   \
                                  (LONG)  0,                               \
                                  (LPSTR) (p1)))


/*----------------------------------------------------------------------+
!                          Function Prototypes                          !
+----------------------------------------------------------------------*/
long Atol(LPSTR x);
void YmdHmsToSystemTime(char *Date, SYSTEMTIME *ST);
long Weekday(SYSTEMTIME *ST);
long TimeFormatter(SYSTEMTIME *SysTime, char *Format, char *Buffer, long BuffLen);

/*----------------------------------------------------------------------+
!                           Global Variables                            !
+----------------------------------------------------------------------*/

/* Superstitious data field to bypass/ignore bugs in certain versions of
!  Windows that cause the beginning of the data area to be overwritten.
!  This field should be the first data area defined and linked. 
*/
char szBugFiller[] = "ABCDEFGHIJKLMNOPQRSTUVWXYZ";
HINSTANCE hinstDLL;                    // DLL Instance saved
long JulianTable[13] = {0,0,31,59,90,120,151,181,212,243,273,304,334};

/* ----------------------------------------------------------------------------+
!   Command Extension Table for this DLL                                       !
!                                                                              !
!   NOTE:   ALL STRINGS MUST BE LOWER CASE.                                    !
!                                                                              !
!   NOTE:   The sizes of the string names CAN NOT exceed 13 characters.        !
!                                                                              !
!   Definitions                                                                !
!     ID       Number, starting from 0, incrementing by 1, terminated by -1.   !
!              For functions, it is also the 'case' number used to access      !
!              the function.                                                   !
!                                                                              !
!     TYPE     Either FUNCTION or CONSTANT.                                    !
!                                                                              !
!     PARAMS/  For constants, this is the actual constant value: long          !
!     VALUE    integer constants only.  For functions, it MUST BE              !
!              0x001F0000 + the number of passed parameters (0 thru 5).        !
!                                                                              !
!     SIZE     Set to 0.                                                       !
!                                                                              !
!     NAME     Name of the function or constant.  MUST BE IN LOWERCASE         !
!              MUST START WITH AN ALPHABETIC CHARACTER.  NO SPACES.            !
!              13 CHARACTERS OR LESS.  USE A NAMING CONVENTION.                !    
!                                                                              !
!   --------------------------------------------------------------------------*/
COMMEXTSTRUCT CommandExtTable[] =
{

/*--------------------------------------------------------------------------*/
/*                                       Name                               */
/* ID   Type      Params/Value  Size  123456789ABCD                         */
/*--------------------------------------------------------------------------*/
  {10,  FUNCTION,  0x001F0001L,  0,  "nowformat" },    // NowFormat(format)
  {11,  FUNCTION,  0x001F0002L,  0,  "timeformat"},    // TimeFormat(ymdhms,format)
/*--------------------------------------------------------------------------*/
  {-1,         0,   0x00000000,  0,  ""             }
};



/*========================================================================+
!                            DllMain                                      !
+========================================================================*/
BOOL WINAPI DllMain(HANDLE hInst, DWORD dwReason, LPVOID lpReserved)
{
      hinstDLL = hInst;
}

/*=======================================================================+
!                             WILExtender                                !
+=======================================================================*/

#ifdef WIN32
DllExport LONG FAR PASCAL WILExtender
#else
LONG FAR PASCAL _export WILExtender
#endif
   (HWND            hWnd,              // handy hWnd for messageboxes, etc
    HINSTANCE       hInst,             // valid hInstance
    LONG            lVersion,          // don't touch
    short           CmdId,             // case number for big switch
    lpAddOnCallback lpPhoneHome,       // don't touch
    LPSTR           lpBlock,           // don't touch
    LONG FAR *      lpLong,            // pointer to your 32 bits
    LPSTR           lp1,               // parameter 1
    LPSTR           lp2,               // parameter 2
    LPSTR           lp3,               // parameter 3
    LPSTR           lp4,               // parameter 4
    LPSTR           lp5)               // parameter 5
{
  char szWork[256];                    // Work Buffer
  
  switch (CmdId)                       // the "big switch"
  {

/****************************************************************************
!          NowFormat(Format)
****************************************************************************/
    case 10:
    {
        SYSTEMTIME Now;
        char *BigBuff;
        long L;    
        
        GetLocalTime(&Now);   
        L = TimeFormatter(&Now, lp1, szWork, 256);
        if (L < 256)
        {
           RETSTRING(szWork);
        }
        /*------------------------------------------------------
        !  Don't know why anyone would need a date string 
        !  bigger than 256 characters,  but if they do
        !  we'll accomodate them.
        +------------------------------------------------------*/
        L += 4;
        BigBuff = (char *) GlobalAlloc(GMEM_FIXED,L);
	    if (BigBuff == NULL)
        {
            RETERROR(-3,3096);
        }
        TimeFormatter(&Now, lp1, BigBuff, L);
        L = ANSWERSTRING(BigBuff);
        GlobalFree(BigBuff);
        return L; 
    }

/*******************************************************************************
*          TimeFormat(Time, Format)
********************************************************************************/
    case 11:
    {
        SYSTEMTIME stm;
        char *BigBuff;
        long L;
    
         YmdHmsToSystemTime(lp1, &stm);

        L = TimeFormatter(&stm, lp2, szWork, 256);
        if (L < 256)
        {
           RETSTRING(szWork);
        }
       
        /*------------------------------------------------------
        !  Don't know why anyone would need a date string 
        !  bigger than 256 characters,  but if they do
        !  we'll accomodate them.
        +------------------------------------------------------*/
        L += 4;
        BigBuff = (char *) GlobalAlloc(GMEM_FIXED,L);
	    if (BigBuff == NULL)
        {
            RETERROR(-3,3096);
        }
        TimeFormatter(&stm, lp2, BigBuff, L);
        L = ANSWERSTRING(BigBuff);
        GlobalFree(BigBuff);
        return L; 
    }


/*******************************************************************************
*  Case 98 is called repeatedly by the WIL extender.  On each call, one        *
*  CommandTable entry is passed back to the WIL extender.  You probably will   *
*  not need to modify this code -- at least at first.                          *
********************************************************************************/
    case 98:                           // Initialization
    {
      LPCOMMEXTSTRUCT lpTable;
      UINT            j;

      lpTable = (LPCOMMEXTSTRUCT) lp1;
      j = (UINT) LOWORD((DWORD) lp2);

      if (CommandExtTable[j].ident == -1)
        return (0);

      lpTable->ident = CommandExtTable[j].ident;
      lpTable->type  = CommandExtTable[j].type;
      lpTable->param = CommandExtTable[j].param;
      lpTable->size  = lstrlen((LPSTR) CommandExtTable[j].delim);

      lstrcpy((LPSTR) lpTable->delim, (LPSTR) CommandExtTable[j].delim);
      return (1);
    }


/********************************************************************************
*   Case 99 is called during the WIL Shutdown process.  If you need to do       *
*   cleanup processing, do it here.                                             *
********************************************************************************/
    case 99:                           // ShutDown
    {
      return (1);
    }

/********************************************************************************
*   Case 100 is used to interrogate the extender.                               *
********************************************************************************/
    case 100:
    {
      LoadStr(1000);
      lstrcpyn(lp1, szWork, (UINT) Atol(lp2));

      return (1);
    }

/********************************************************************************
*                                default case                                   *
********************************************************************************/
       default:                           // we shouldn't get here
    {
      RETERROR(-1, 1001);              // unknown function
      lp1 = lp2 = lp3 = lp4 = lp5 = NULL;
      lpLong = 0;
    }
  }                                    // end of "big switch"
}









/*************************************************************
 *************************************************************
 **                                                         **
 **     ********************************************        **
 **     **                                        **        **
 **     **   Common Routines                      **        **
 **     **                                        **        **
 **     ********************************************        **
 **                                                         **
 *************************************************************
 *************************************************************/








/********************************************************************************
*                                                                               *
*  Given a date in the WinBatch format YY:MM:DD:HH:MM:SS                        *
*  Convert it to a SystemTime block.                                            *
*                                                                               *
*   Use of four digit years is recommended.                                     *
*   Two digit years are assumed to be 20xx.                                     *
*                                                                               *
********************************************************************************/
void YmdHmsToSystemTime(char *Date, SYSTEMTIME *ST)
{
    long Y, M, D, X;
    char C;
    
    Y = 0;
    while (1) {C=*Date; if (C < '0' || C > '9') break; Y=(Y*10)+(C-'0'); Date++;}
    if (Y < 100) Y += 2000;
    ST->wYear   = Y;

    M = 0;
    if (*Date == ':') Date++;
    while (1) {C=*Date; if (C < '0' || C > '9') break; M=(M*10)+(C-'0'); Date++;}
    ST->wMonth  = M; 

    D = 0;
    if (*Date == ':') Date++;
    while (1) {C=*Date; if (C < '0' || C > '9') break; D=(D*10)+(C-'0'); Date++;}
    ST->wDay  = D;    
    
    X = 0;
    if (*Date == ':') Date++;
    while (1) {C=*Date; if (C < '0' || C > '9') break; X=(X*10)+(C-'0'); Date++;}
    ST->wHour  = X;
    
    X = 0;
    if (*Date == ':') Date++;
    while (1) {C=*Date; if (C < '0' || C > '9') break; X=(X*10)+(C-'0'); Date++;}
    ST->wMinute  = X;

    X = 0;
    if (*Date == ':') Date++;
    while (1) {C=*Date; if (C < '0' || C > '9') break; X=(X*10)+(C-'0'); Date++;}
    ST->wSecond  = X;
    
    ST->wMilliseconds = 0;
    ST->wDayOfWeek = Weekday(ST);

    return;
}


/********************************************************************************
*                                                                               *
*    Compute the day of the week for a date.  0=Sun, 1=Mon, 2=Tue ...           *
*    Uses the "Zeller Congruence".                                              *
*                                                                               *
********************************************************************************/ 
long Weekday(SYSTEMTIME *TimeWork)
{
    long I;
    long MM, YY;
    
    MM = TimeWork->wMonth;
    YY = TimeWork->wYear;

    if (MM < 3) {MM += 12; YY -= 1;}

    I = (TimeWork->wDay + ((26*(MM+1))/10) + ((125*YY)/100) - (YY/100) + (YY/400)) %7 ;
    if (I == 0) I=7;
    I--;
    return I;    
}
       
/******************************************************************************
*                                                                             *
*   Convert a decimal character string to a long integer.                     *
*                                                                             *
******************************************************************************/

long Atol(LPSTR lpSource)
{
    LONG  nNum;
    BOOL  bNegate;

    bNegate = FALSE;
    nNum = 0;

    while (*lpSource == ' ')   lpSource++;
    if (*lpSource == '-' || *lpSource == '+')
    {
       if (*lpSource == '-')  bNegate = TRUE;
       lpSource++;
    }
    while (*lpSource >= '0' && *lpSource <= '9')
    {
       nNum = nNum * 10 + (*lpSource - '0');
       lpSource++;
    }
    if (bNegate)  nNum = -nNum;
    return (nNum);
}

/*************************************************************
!   Macros used by the TimeFormatter function.
!
!   LtoA2 converts a long to a two-digit character string.
!   LtoA4 converts a long to a four-digit character string.
!   LtoA4z converts a long to a two digit string, zero-suppressed.
!
!   comp2 compares a two character string to the Format
!   comp3 compares a three character string to the Foramt
!   comp4 compares a four character strin to the Format
!
!   Note- before using any of the comp routines we have 
!   already tested the first character,  so the routines
!   only test the remaining bytes.
**************************************************************/
#define LtoA2(p1)    szWork[0]=((p1/10)%10)+'0';szWork[1]=(p1%10)+'0';szWork[2]='\0'
#define LtoA4(p1)    szWork[0]=((p1/1000)%10)+'0';szWork[1]=((p1/100)%10)+'0';szWork[2]=((p1/10)%10)+'0';szWork[3]=(p1%10)+'0';szWork[4]='\0'
#define LtoA2z(p1)   {long Ix=0; if (p1>9) {szWork[0]=((p1/10)%10)+'0';Ix=1;} szWork[Ix]=(p1%10)+'0';Ix++;szWork[Ix]='\0';}

#define comp2(S) *(Format+1)==S[1]
#define comp3(S) *(Format+1)==S[1]&&*(Format+2)==S[2]
#define comp4(S) *(Format+1)==S[1]&&*(Format+2)==S[2]&&*(Format+3)==S[3]


/****************************************************************************
*                                                                           *
*  Given a date and a format,                                               *
*  convert the date to a string as directed by the format.                  *
*                                                                           *
*  If the buffer is too short to receive the entire formatted date          *
*  we return the portion that fits.  The return value is the                *
*  actual length required.                                                  *
*                                                                           *
****************************************************************************/        
long TimeFormatter(SYSTEMTIME *SysTime, char *Format, char *Buffer, long BuffLen)
{
    char szWork[128];
    char Repeat;
    char jjj[4];
    long H12;
    long K;
    char C;
    char *Cx;
    long DateLength;
       

    DateLength = 0;
    
          /*----------------------------------------------------------+
          !      Compute the julian day                               !
          +----------------------------------------------------------*/  
    K = JulianTable[SysTime->wMonth] + SysTime->wDay;
    if (SysTime->wMonth > 2)
      if (SysTime->wYear % 4 == 0)
        if (SysTime->wYear % 100 != 0 || SysTime->wYear % 400 == 0)
            K += 1;
    
    jjj[0] = ((K / 100) % 10)+ '0';
    jjj[1] = ((K / 10) % 10) + '0';    
    jjj[2] = (K % 10) + '0';
    jjj[3] = '\0';         

          /*----------------------------------------------------------+
          !      Compute the day in 12-hour format                    !
          +----------------------------------------------------------*/  
    H12 = (SysTime->wHour);
    if (H12 == 0) H12 = 12;
    if (H12 > 12) H12 -= 12;

          /*----------------------------------------------------------+
         !  Copy the format to the buffer, looking for substititutions 
         +-----------------------------------------------------------*/
     while (1)
     {
         C = *Format;                    // Next character.
         if (C == '\0') break;           // At end of format?                              
      
         switch (C)
         {
            case 'a':
                if (comp2("am")) 
                {  
                    Format+=2;
                    if (SysTime->wHour < 12) LoadStr(401);
                    else                     LoadStr(402);
                    break;
                 }
            
             case 'A':
                if (comp2("AM"))         
                {  
                    Format+=2; 
                    if (SysTime->wHour < 12) LoadStr(403);
                    else                     LoadStr(404);
                    break;
                 }

             case 'Y':
                if (comp4("YYYY")) {Format+=4; LtoA4(SysTime->wYear); break;}
                if (comp2("YY"))   {Format+=2; LtoA2(SysTime->wYear); break;}
                
             case 'M':
                if (comp4("MMMM")) {Format+=4; LoadStr(300+SysTime->wMonth);break;} 
                if (comp3("MMM"))  {Format+=3; LoadStr(200+SysTime->wMonth);break;}
                if (comp2("MM"))   {Format+=2; LtoA2(SysTime->wMonth);      break;}
                
             case 'D':
                if (comp4("DDDD")) {Format+=4; LoadStr(110+SysTime->wDayOfWeek);break;}
                if (comp3("DDD"))  {Format+=3; LoadStr(100+SysTime->wDayOfWeek);break;}
                if (comp2("DD"))   {Format+=2; LtoA2(SysTime->wDay);            break;}

             case 'H':  
                if (comp2("HH"))   {Format+=2; LtoA2(SysTime->wHour); break;}
 
             case 'h':  
                if (comp2("hh"))   {Format+=2; LtoA2(H12); break;}
                   
             case 'm': 
                if (comp2("mm"))   {Format+=2; LtoA2(SysTime->wMinute); break;}


             case 's':
                if (comp2("ss"))   {Format+=2; LtoA2(SysTime->wSecond); break;}
    
             case 'Z':
                if (comp2("ZM"))   {Format+=2; LtoA2z(SysTime->wMonth); break;}          
                if (comp2("ZD"))   {Format+=2; LtoA2z(SysTime->wDay);   break;}
                if (comp2("ZH"))   {Format+=2; LtoA2z(SysTime->wHour);  break;} 

             case 'z':
                if (comp2("zh"))   {Format+=2; LtoA2z(H12);break;}
                if (comp2("zm"))   {Format+=2; LtoA2z(SysTime->wMinute);break;}
                if (comp2("zs"))   {Format+=2; LtoA2z(SysTime->wSecond);break;}  
                           
             case 'J':
                if (comp3("JJJ"))  {Format+=3; lstrcpy(szWork,jjj); break;}

             default:
                Format++;
                DateLength++;
                if (DateLength < BuffLen) *Buffer++ = C;
                continue;
         }

         Cx = szWork;
         while (1)
         {
            C = *Cx++;
            if (C == '\0') break;
            DateLength++;
            if (DateLength < BuffLen) *Buffer++ = C;
         }

     } 
     *Buffer = '\0';              // Terminate the buffer;
     return DateLength;           // Return Estimated Length.
}

